import * as https from "https";
import * as http from "http";
import * as fs from "fs";
import { createHash } from "node:crypto";
import { Parser } from "json2csv";
import { parse } from "url";
import HTMLParser from "node-html-parser";
import HTMLElement from "node-html-parser/dist/nodes/html";
const dataFolder = "data/";

const username = process.env.USERNAME;
const password = process.env.PASSWORD;
if (!username || !password) throw "auth not set";
const auth = Buffer.from(username + ":" + password).toString("base64");

const json2csvParser = new Parser();
const interval = 5000;
let working = false;
type dataType = Exclude<Awaited<ReturnType<typeof getMeeting>>, false>;
if (!fs.existsSync(dataFolder)) fs.mkdirSync(dataFolder);
let requestCache = Promise.resolve("");
const getHash = (url: string) =>
  dataFolder + createHash("md5").update(url).digest("hex");
function getFile(url: string, disableCache = false) {
  const filename = getHash(url);
  let data = "";
  if (!disableCache) {
    try {
      data = fs.readFileSync(filename, "utf8");
    } catch (_) {
      //do nothing :)
    }
  }
  return data;
}

async function get(url: string, disableCache = false) {
  let data = getFile(url, disableCache);
  if (!data) {
    const options = {
      hostname: "www.swimrankings.net",
      path: "/index.php" + url,
      headers: {
        "User-Agent":
          "swimrankings.rubend.nl or zwemtijden.rubend.nl contact ruben at rubend.nl",
        Authorization: "Basic " + auth,
      },
    };
    requestCache = requestCache
      .then(() => new Promise((resolve) => setTimeout(resolve, interval)))
      .then(
        () =>
          new Promise((resolve) =>
            https.get(options, (res) => {
              let data = "";
              res.on("data", (chunk) => {
                data += chunk;
              });
              res.on("end", () => {
                const filename = getHash(url);
                fs.writeFileSync(filename, data, "utf8");
                resolve(data);
              });
            }),
          ),
      );
    data = await requestCache;
  }
  return HTMLParser.parse(data);
}
async function getAthlete(athleteId: number) {
  const root = await get(
    `?page=athleteDetail&athleteId=${athleteId}&athletePage=MEET`,
    true,
  );
  return {
    urls: root
      .querySelectorAll(".athleteMeet0,.athleteMeet1")
      .map((item) => item.querySelector("a")!.getAttribute("href")!)
      .filter((url) => url.startsWith("?page")),
    fina: root
      .querySelector('[name="points"] > [selected]')!
      .getAttribute("value")!
      .split("_")[1],
  };
}
async function getMeeting(athleteId: number, url: string) {
  const root = await get(url);

  if (root.innerHTML.length < 500) return false;
  const titleLeft = root.querySelectorAll(".titleLeft");
  const titleRight = root.querySelectorAll(".titleRight");
  const meetData = {
    meetName: titleLeft[0].text,
    city: titleLeft[1].text,
    date: titleRight[1].text,
  };
  const items: HTMLElement[] = [];
  let topItem = root.querySelector(
    `.nameImportant > a[href="?page=athleteDetail&athleteId=${athleteId}"]`,
  )!.parentNode!.parentNode!.nextElementSibling;
  while (topItem?.classNames.includes("meetResult0")) {
    items.push(topItem);
    topItem = topItem.nextElementSibling;
  }
  return items
    .map((item) => item.childNodes.map((item) => item.text))
    .map((item) => {
      const [
        stroke,
        heatType,
        rawRanking,
        time,
        points,
        poolType,
        //previousTime,
        //previousPercentage,
        //previousDate,
      ] = item;
      let ranking = rawRanking.split(".")[0];
      if (ranking[0] == " ") ranking = ranking.substring(1);
      return {
        stroke,
        heatType,
        ranking,
        time,
        points,
        poolType,
        ...meetData,
      };
    });
}
function tableToHtml(data: dataType) {
  let html = "<table>";
  html += `<tr>${Object.keys(data[0])
    .map((item) => `<th>${item}</th>`)
    .join("")}</tr>`;
  html += data
    .map(
      (line) =>
        `<tr>${Object.values(line)
          .map((item) => `<td>${item}</td>`)
          .join("")}</tr>`,
    )
    .join("");
  html += "</table>";
  return html;
}
http
  .createServer(async (req, res) => {
    if (working) {
      res.end("text\nServer currently busy.");
      return;
    }
    const url = parse(req.url!, true);
    const fileType = url.pathname!.split("/")[1];
    if (!["json", "csv", "html"].includes(fileType)) {
      res.end("invalid file type!");
      return;
    }
    const query = url.query;
    if (query.id) {
      const swimmerId = +query.id;
      const athlete = await getAthlete(swimmerId);
      const fina = query.fina || athlete.fina;
      const data: dataType = [];
      const urls = athlete.urls.map((url) => url + "&points=fina_" + fina);
      const afterMap = urls.map((url) => getFile(url));
      const toGet = afterMap.filter((data) => !data).length;
      const delay = toGet * interval;
      const willFinish = delay <= 29000; //29 seconds, 30 is the max for R
      if (!willFinish && query.requestedBy == "zwemtijden") {
        let seconds = (delay * 1.1) / 1000;
        let minutes = Math.floor(seconds / 60);
        const hours = Math.floor(minutes / 60);
        minutes = Math.floor(minutes % 60);
        seconds = Math.floor(seconds % 60);
        const time = `${hours}:${minutes}:${seconds}`;
        res.end(
          "text\nPlease wait at least " +
            time +
            " before requesting again. Warning: this time will not be shown again.",
        );
      }
      working = true;
      for (const url of urls) {
        const response = await getMeeting(swimmerId, url);
        if (response) data.push(...response);
      }
      working = false;
      if (!willFinish) return;
      res.setHeader(
        "Content-Disposition",
        `inline; filename="${swimmerId}.${fileType}"`,
      );
      if (fileType == "json") {
        res.setHeader("Content-Type", "application/json");
        res.end(JSON.stringify(data));
      }
      if (fileType == "csv") {
        res.setHeader("Content-Type", "text/csv");
        res.end(json2csvParser.parse(data));
      }
      if (fileType == "html") {
        res.setHeader("Content-Type", "text/html; charset=utf-8");
        res.end(tableToHtml(data));
      }
    }
  })
  .listen(8080);
